<html>
    <head>
        <title>GA4GH reference server {{ info.getServerVersion()}}</title>
        <link rel="shortcut icon" href="{{ url_for('static', filename='favicon.ico') }}">
    </head>
    <body>
        <h2>GA4GH reference server {{ info.getServerVersion() }}</h2>
        <div>
            <h3>Protocol version {{ info.getProtocolVersion() }}</h3>
        </div>
        <div>
            <h3>Operations available</h3>
            <table>
                <th>Method</th>
                <th>Path</th>
                {% for rule in info.getUrls() %}
                <tr>
                    <td>{{ rule[1] }}</td>
                    <td>{{ rule[0] }}</td>
                </tr>
                {% endfor %}
            </table>
        </div>
        <div>
            <h3>Uptime</h3>
            Running since {{ info.getNaturalUptime()}} ({{ info.getPreciseUptime()}})
        </div>
        <div>
            <h3>Configuration</h3>
            <table>
                <th>Key</th>
                <th>Value</th>
                {% for key, value in info.getConfiguration() %}
                <tr>
                    <td>{{ key }}</td>
                    <td>{{ value }}</td>
                </tr>
                {% endfor %}
            </table>
        </div>
        <div>
            <h3>Data</h3>
            <h4>VariantSets</h4>
            <table>
                {% for variantSet in info.getVariantSets() %}
                <tr>
                    <td>{{ variantSet.getId() }}</td>
                </tr>
                {% endfor %}
            </table>
            <h4>ReadGroupSets</h4>
            <table>
                {% for readGroupSet in info.getReadGroupSets() %}
                <tr>
                    <td>{{ readGroupSet.getId() }}</td>
                </tr>
                    {% for readGroup in readGroupSet.getReadGroups() %}
                <tr>
                    <td></td>
                    <td>{{ readGroup.getId() }}</td>
                </tr>
                {% endfor %}
            {% endfor %}
            </table>
        </div>
    </body>
</html>
